__author__ = "Fabio Caccamo"
__copyright__ = "Copyright (c) 2019-present - Fabio Caccamo"
__description__ = (
    "encode / decode italian fiscal codes "
    "- codifica / decodifica del Codice Fiscale italiano."
)
__email__ = "fabio.caccamo@gmail.com"
__license__ = "MIT"
__title__ = "codicefiscale"
__version__ = "0.7.0"
